/*
    learn.cpp
    The Latent Words Language Model, Gibbs Sampler.
    $Id: learn.cpp,v 1.13 2010/03/19 10:16:54 daichi Exp $

*/
#include "lwlm.h"
#include "hpylm.h"
#include "ngram.h"
#include "random.h"
#include "util.h"
#include "debug.h"
#include "log.h"
#include <cstdio>
#include <cstdlib>
#include <vector>
#include <algorithm>

typedef struct {
	double slice;
	ngram *context;
	vector<int> *agenda;
} sparam;

static vector<int> seen;

static int push_agenda_if (restaurant *rest, sparam *param);

bool
LWLM::learn ()
{
	witerator hidden, observe;
	int i, j, k, width, nwords;
	int start, elapsed;
	siterator s;
	double ppl;

	/* prepare */
	lm->lexicon = L;
	table->init (L);
	eprintf("vocabulary = %d\n", L);

	/* parse */
	for (s = observed.begin(), j = 0; s != observed.end(); s++, j++)
	{
		parse (*s);
		if (j % 1000 == 0)
			eprintf("parsing %5d/%5d ..\r", j, observed.size());
	}
	eprintf("parsing %5d/%5d ..done.\n", j, observed.size());
	
	start = myclock();
	nwords = words.size();

	/* Gibbs */
	for (j = 0; j < iter; j++)
	{
		lm->hyper_estimate ();
		
		random_shuffle (words.begin(), words.end(), irand);

		for (i = 0; i < nwords; i++)
		{
			width = words[i]->width;
			hidden = words[i]->hidden;
			observe = words[i]->observed;

			for (k = width - 1; k >= 0; k--) {
				if (lm->remove_customer (hidden + k) == false)
				{
					error("remove_customer failed!!\n");
					exit(1);
				}
			}
			if (j > 0)
				table->sub_count (*hidden, *observe);
			{
				*hidden = slice_gibbs_draw (hidden, observe, width);
			}
			table->add_count (*hidden, *observe);
			for (k = 0; k < width; k++) {
				if (lm->add_customer (hidden + k) == false)
				{
					error("add_customer failed!!\n");
					exit(1);
				}
			}
			/*
			 *  ETA
			 *
			 */
			if ((i + 1) % 100 == 0)
			{
				elapsed = myclock() - start + 1;
				eprintf("[%2d] : sampling %6d/%6d.. ETA: %s (%d words/sec)\r",
					j + 1, i + 1, nwords,
					rtime(elapsed * (iter / (j + i/(double)nwords) - 1)),
					int((nwords * j + i + 1)/(double)elapsed));
			}
			
		}
		ppl = perplexity();
		lprintf("Gibbs [%d] : PPL = %g\n", j + 1, ppl);
		/*
		 *  ETA
		 *
		 */
		elapsed = myclock() - start;
		eprintf("[%2d] : sampling %6d/%6d.. ETA: %s (%d words/sec) PPL = %g\r",
			j + 1, i + 1, nwords,
			rtime(elapsed * (iter / (j + i/(double)nwords) - 1)),
			int((nwords * j + i + 1)/(double)elapsed + 0.5),
			ppl);
	}
	eprintf("\ndone.\n");

}

int
LWLM::slice_gibbs_draw (int *hidden, int *observed, int width)
{
	return slice_gibbs_draw ((witerator)hidden, (witerator)observed, width);
}

int
LWLM::slice_gibbs_draw (witerator hidden, witerator observed, int width)
{
	static vector<int> agenda;
	static vector<double> buf;
	ngram *context;
	witerator w;
	double p;
	int k;
	
	agenda.resize (0);
	buf.resize (0);

	/* draw candidates */
	if (slice_sample_agenda (&agenda, hidden, hidden - 1))
	{
		if (agenda.size() == 1)
			return agenda[0];
	}
	else {
		error("slice_gibbs_draw: agenda invalid.");
		return -1;
	}

	/* probability */
	for (w = agenda.begin(); w != agenda.end(); w++)
	{
		*hidden = *w;

		for (k = 1, p = 1; k < width; k++)
		{
			p *= lm->hpy_probability (hidden + k, hidden + k - 1);
		}
		p *= table->probability (*hidden, *observed);
		
		buf.push_back (p);
	}

	if ((k = draw_discrete(&buf[0], buf.size())) == -1)
	{
		error("slice_gibbs_draw: sampling failed.");
	}

	return agenda[k];
}
	

bool
LWLM::slice_sample_agenda (vector<int> *agenda, witerator word, witerator history)
{
	double p, slice;
	ngram *context;

	if ((p = lm->hpy_probability(word, history)) > 0)
	{
		slice = RANDOM * p;
	}
	else {
		error("slice_sample_agenda: invalid slice. p = %g, slice = %g\n",
		      p, slice);
		return false;
	}

	if ((context = lm->find_context(history)) == NULL)
	{
		error("slice_sample_agenda: can't find the context.\n");
		return false;
	}

	if (seen.empty())
		seen.resize (L);
	fill (seen.begin(), seen.end(), 0);

	hpylm_slice_agenda (context, slice, agenda);

	if (agenda->size() == 0)  /* no candidate sampled */
	{
		agenda->push_back (*word);
	}

	return true;
}

bool
LWLM::hpylm_slice_agenda (ngram *node, double slice, vector<int> *agenda)
{
	static sparam param;
	record *records = node->records;
	double residual = node->hpy_residuals();

	param.slice = slice;
	param.agenda = agenda;
	param.context = node;

	records->foreach ((recfunc)push_agenda_if, &param);

	if (slice < residual)
	{
		if (node->parent != NULL)
			return hpylm_slice_agenda (node->parent, slice / residual, agenda);
	}

}

static int
push_agenda_if (restaurant *rest, sparam *param)
{
	double p, slice = param->slice;
	ngram *context = param->context;
	vector<int> *agenda = param->agenda;

	if (rest->word == EOS)
	{
		return 0;	/* exclude EOS */
	}

	p = context->hpy_probability (rest->word);

	if (p > slice)
	{
		if (!seen[rest->word])
		{
			agenda->push_back (rest->word);
			seen[rest->word] = 1;
		}
		return 0;
	}
	else {
		return 1;    /* prune afterwards */
	}
}

int       
LWLM::gibbs_draw (int *hidden, int *observed, int width)
{
	return gibbs_draw ((witerator)hidden, (witerator)observed, width);
}

int       
LWLM::gibbs_draw (witerator hidden, witerator observed, int width)
{
	static vector<double> p;
	ngram *context;
	int w, k;

	if (p.empty())
		p.resize (L);

	if ((context = lm->find_context (hidden - 1)) == NULL)
	{
		error("gibbs_draw: can't find context.\n");
		return -1;
	}
	for (w = 0; w < L; w++)
		p[w] = context->hpy_probability (w);

	for (w = 0; w < L; w++)
	{
		*hidden = w;
		for (k = 1; k < width; k++)
			p[w] *= lm->hpy_probability (hidden + k, hidden + k - 1);
		p[w] *= table->probability (*hidden, *observed);
	}
	
	if ((w = draw_discrete(&p[0], p.size())) == -1)
	{
		error("draw_gibbs: sampling failed.\n");
	}

	return w;
}

