/*
    load.cpp
    $Id: load.cpp,v 1.7 2010/03/18 07:28:40 daichi Exp $

*/
#include <iostream>
#include <fstream>
#include <string>
#include <utility>
#include <ext/hash_map>
#include <cstdio>
#include <cstring>
#include "lwlm.h"
#include "ngram.h"
#include "util.h"
#include "io.h"

using namespace std;

static char whitespace[] = " \t\n";

node_map_rev *
hpylm_load_nodes (char *file, HPYLM *lm)
{
	node_map_rev *hash = new node_map_rev;
	ifstream ifs (strconcat(file, "/nodes"));
	restaurant *rest;
	ngram *node;
	char *token;
	string line;
	int j, l, seed = 0;

	if (ifs.fail())
	{
		perror("ifstream");
		return NULL;
	}

	for (l = 0; getline(ifs, line);)
	{
		if (isspaces(line)) {
			l = 0;
			continue;
		}

		if (l == 0)		/* node */
		{
			if ((node = new ngram) == NULL)
			{
				error("hpylm_load_nodes: can't create a node.\n");
				return NULL;
			}
			hash->insert (cons(seed++, node));
			
			if (sscanf(line.c_str(), "%d\t%d\t%d\t%d\n",
				   &node->level,
				   &node->word,
				   &node->ncounts,
				   &node->ntables) == 0)
			{
				error("hpylm_load_nodes: node information error.\n");
				error(line.c_str());
				return NULL;
			}
		} else if (l == 1)	/* bounds */
		{
			for (token = strtok((char *)line.c_str(), whitespace); token;
			     token = strtok(NULL, whitespace))
			{
				node->records->bounds.push_back (atoi(token));
			}
		}
		else {			/* restaurant */
			if ((rest = new restaurant(node)) == NULL)
			{
				error("hpylm_load_nodes: can't create a restaurant.\n");
				return NULL;
			}
			for (token = strtok((char *)line.c_str(), whitespace), j = 0; token;
			     token = strtok(NULL, whitespace), j++)
			{
				switch (j) {
				case 0: rest->word = atoi(token); break;
				case 1: rest->ncounts = atoi(token); break;
				default:
					rest->tables.push_back (atoi(token));
				}
			}
			/* register */
			insert_word (node->words, rest);
			node->records->v.push_back (rest);
		}
		l++;
	}
			  
	ifs.close();
	return hash;
}

ngram *
hpylm_load_trees (char *file, HPYLM *lm, node_map_rev *nodemap)
{
	ifstream ifs (strconcat(file, "/tree"));
	ngram *node, *child, *tree;
	string line;
	int id, l;

	if (ifs.fail())
	{
		perror("ifstream");
		return NULL;
	}

	for (l = 0; getline(ifs, line);)
	{
		if (isspaces(line)) {
			l = 0;
			continue;
		}
		if (sscanf(line.c_str(), "%d", &id) == 0)
		{
			error("hpylm_load_trees: invalid line.\n");
			return NULL;
		}

		if (l == 0)	/* parent node */
		{	
			node = (*nodemap)[id];
			if (node->level == 0) {
				tree = node;
			}
		} else {	/* child node */
			child = (*nodemap)[id];
			child->parent = node;
			insert_node (node->children, child);
		}
		l++;
	}

	ifs.close();
	return tree;
}

void
hpylm_load_param (char *file, HPYLM *lm)
{
	FILE *fp;
	int n;

	if ((fp = fopen(strconcat(file, "/param"), "r")) == NULL)
	{
		perror("fopen");
		return;
	}

	if (fscanf(fp, "%d\t%d", &lm->order, &lm->lexicon) != 2)
	{
		error("hpylm_load_param: invalid line.\n");
		return;
	}
	for (n = 0; n <= lm->order; n++)
	{
		if (fscanf(fp, "%lg\t%lg", &lm->d[n], &lm->theta[n]) != 2)
		{
			error("hpylm_load_param: invalid line.\n");
			return;
		}
	}

	fclose(fp);
}

bool
lwlm_load_param (char *model, int *order, double *alpha)
{
	FILE *fp;

	if ((fp = fopen(strconcat(model, "/param"), "r")) == NULL)
	{
		error("lwlm_load_param: can't read %s.\n", model);
		return false;
	}
	if (fscanf(fp, "%d\t%lg", order, alpha) != 2)
	{
		error("lwlm_load_param: can't load parameters.\n");
		return false;
	}
	fclose(fp);

	return true;
}

void
HPYLM::load (char *file)
{
	node_map_rev *nodemap;

	if ((nodemap = hpylm_load_nodes (file, this)) == NULL)
	{
		error("HPYLM::load: can't load nodes.\n");
		return;
	}
	if ((tree = hpylm_load_trees (file, this, nodemap)) == NULL)
	{
		error("HPYLM::load: can't load trees.\n");
		return;
	}
	hpylm_load_param (file, this);

	delete nodemap;
}


void
LWLM::load_lexicon (char *file)
{
	ifstream ifs (strconcat(file, "/lexicon"));
	string sline;
	char *s;

	if (ifs.fail())
	{
		error("load_lexicon: can't find file %s.\n", file);
		return;
	}

	while (getline(ifs, sline))
	{
		s = (char *)sline.c_str();
		sid (s);
	}
	eprintf("vocabulary = %d\n", L);
	
	ifs.close();
}
	
void
LWLM::load (char *model)
{
	load_lexicon (model);
	eprintf("loading LM..\n");
	lm->load (strconcat(model, "/lm"));
	eprintf("loading table..\n");
	table->init (L);
	table->load (model, this);
}

