/*
    LWLM.cpp
    the Latent Words Language Model.
    $Id: lwlm.cpp,v 1.11 2010/03/19 10:16:54 daichi Exp $

*/
#include "lwlm.h"
#include "hpylm.h"
#include "util.h"
#include <iostream>
#include <fstream>
#include <string>
#include <utility>
#include <cstring>
#include <cstdio>
#include <cmath>
#include <ext/hash_map>

using namespace std;
using namespace __gnu_cxx;

static const char whitespace[] = " \t";

LWLM::LWLM (int order_, double alpha_, int iter_)
{
	order = order_;
	alpha = alpha_;
	iter  = iter_;
	L     = 0;
	lm    = new HPYLM (order);
	table = new TransTable (alpha);
	eprintf("latent ngram order = %d, alpha = %g\n", order + 1, alpha);
}

int
LWLM::sid (char *s)
{
	lex_hash::iterator it;
	string word = s;

	if ((it = dict.find(word)) == dict.end())
	{
		dict.insert (cons(word, L));
		rdict.insert (cons(L, word));
		return L++;
	} else {
		return it->second;
	}
}

char *
LWLM::ids (int id)
{
	id_hash::iterator it;

	if (id == EOS)
		return "<EOS>";

	if ((it = rdict.find(id)) == rdict.end())
	{
		return "<null>";
	} else {
		return (char *) it->second.c_str();
	}
}

bool
LWLM::read (char *file)
{
	int n;
	string buf;
	siterator it;
	vector<int> *s;
	char *line, *token;
	ifstream ifs (file);

	if (ifs.fail())
	{
		error("can't open %s to parse.\n", file);
		return false;
	}

	eprintf("reading %s.. ", file);
	while (getline(ifs, buf))
	{
		line = (char *)buf.c_str();

		if (isspaces(line))
			continue;
		else
			s = new vector<int>;

		/* pad */
		for (n = 0; n < order; n++)
		{
			s->push_back (EOS);
		}
		/* parse */
		for (token = strtok(line, whitespace); token;
		     token = strtok(NULL, whitespace))
		{
			s->push_back (sid(token));
		}
		/* EOS */
		s->push_back (EOS);
		
		/* register */
		observed.push_back (s);
	}
	eprintf("done.\n");
}

bool
LWLM::parse (vector<int> *s)
{
	vector<int> *h;
	witerator ot, ht;
	int t, width, size = s->size();

	/* copy observed */
	if ((h = new vector<int>(*s)) != NULL)
	{
		hidden.push_back (h);
	}

	/* sample points */
	for (t = order; t < size - 1; t++)	/* exclude EOS from points */
	{
		ot = s->begin() + t;
		ht = h->begin() + t;
		if (t + order < size) {
			width = order + 1;
		} else {
			width = size - t;
		}
		words.push_back (new Word (ot, ht, width));
	}
	
	/* n-gram */				/* include EOS */
	for (ht = h->begin() + order; ht != h->end(); ht++)
	{
		lm->add_customer (ht);
	}
}

double
LWLM::perplexity ()
{
	int i, t, n = 0, m = observed.size();
	double l = 0;

	for (i = 0; i < m; i++)
	{
		vector<int> *h = hidden[i], *o = observed[i];
		for (t = order; t < h->size() - 1; t++)
		{
			l += log (lm->hpy_probability(&(*h)[t], &(*h)[t - 1]))
				+ log (table->probability((*h)[t], (*o)[t]));
			n++;
		}
	}

	return exp(- l / n);
}

void
LWLM::print ()
{
	siterator s;
	witerator it;

	for (s = observed.begin(); s != observed.end(); s++)
	{
		for (it = (*s)->begin(); it != (*s)->end(); it++)
			printf("%d ", *it);
		printf("\n----\n");
	}
	
}


