/*
    ngram.cpp
    HPYLM language model class.
    $Id: ngram.cpp,v 1.6 2010/03/19 10:16:54 daichi Exp $

*/
#include "ngram.h"
#include "hpylm.h"
#include "random.h"
#include "record.h"
#include "hash.h"
#include "util.h"
#include "debug.h"
#include <vector>
#include <cassert>
#include <cstdlib>

using namespace std;

HPYLM *ngram::lm;

/*
 *   ngram
 *
 */

ngram::ngram (ngram *parent_, int word_, HPYLM *lm_)
{
	lm       = lm_;
	word     = word_;
	parent   = parent_;
	children = new hashmap;
	words    = new hashmap;
	records  = new record;
	ncounts  = 0;
	ntables  = 0;
	if (parent == NULL)
		level = 0;
	else
		level = parent->level + 1;
	{
		words->set_deleted_key (EOS - 1);
		children->set_deleted_key (EOS - 1);
#ifndef MEMORY
 		words->set_empty_key (EOS - 2);
 		children->set_empty_key (EOS - 2);
#endif
	}
}	

ngram::ngram (ngram *parent_, int word_)
{
	word     = word_;
	parent   = parent_;
	words    = new hashmap;
	children = new hashmap;
	records  = new record;
	ncounts  = 0;
	ntables  = 0;
	if (parent == NULL)
		level = 0;
	else
		level = parent->level + 1;
	{
		words->set_deleted_key (EOS - 1);
		children->set_deleted_key (EOS - 1);
#ifndef MEMORY
		words->set_empty_key (EOS - 2);
		children->set_empty_key (EOS - 2);
#endif
	}
}

ngram::ngram ()
{
	parent   = NULL;
	words    = new hashmap;
	children = new hashmap;
	records  = new record;
	ncounts  = 0;
	ntables  = 0;
	{
		words->set_deleted_key (EOS - 1);
		children->set_deleted_key (EOS - 1);
#ifndef MEMORY
 		words->set_empty_key (EOS - 2);
 		children->set_empty_key (EOS - 2);
#endif
	}
}


ngram::~ngram ()
{
	assert(ncounts == 0);
	assert(ntables == 0);

	delete_node (parent->children, this);
	delete records;
	delete children;
	delete words;
}

bool
ngram::increment_customer (int w)
{
	restaurant *rest;

	if ((rest = find_word (words, w)) == NULL)
	{
		if ((rest = new restaurant (w, this)) == NULL)
		{
			error("increment_customer: can't create a restaurant %d.\n", w);
			return false;
		}
	}
	return rest->increment_customer();
}

bool
ngram::decrement_customer (int w)
{
	restaurant *rest;

	if ((rest = find_word (words, w)) == NULL)
	{
		error("decrement_customer: can't find a word %d.\n", w);
		return false;
	} 
	return rest->decrement_customer();
}


/*
 *  restaurant
 *
 */

restaurant::restaurant (int word_, ngram *context_)
{
	word    = word_;
	context = context_;
	ncounts = 0;
	cur     = context->records->add (this);
	/* register */
	insert_word (context->words, this);
}

restaurant::restaurant (ngram *context_)
{
	context = context_;
}

restaurant::~restaurant ()
{
	assert (ncounts == 0);
	delete_word (context->words, this);
}

/* increment */
	
bool
restaurant::increment_customer ()
{
	static vector<double> buf;

	HPYLM *lm = context->lm;
	ngram *parent = context->parent;
	int ntables = tables.size();
	int l = context->level;
	int k, i;

	/* allocate */
	if (ntables >= buf.size())
		buf.resize (ntables + 1);

	/* choose */
	if (ntables > 0)
	{
		/* PY discounts */
		for (i = 0; i < ntables; i++)
			buf[i] = tables[i] - lm->d[l];
		if (parent != NULL) {
			buf[ntables] = (lm->theta[l] + lm->d[l] * context->ntables)
				* parent->hpy_probability (word);
		}
		else {
			buf[ntables] = (lm->theta[l] + lm->d[l] * context->ntables)
				/ lm->lexicon;
		}

		/* draw k */
		if ((k = draw_discrete(&(buf[0]), ntables + 1)) == -1)
		{
			error("increment_customer: table invalid. (ntables = %d)\n", ntables);
			return false;
		}
	} else {
		k = 0;
	}

	/* add or increment a table */
	if (k < ntables) {
		increment_table (k);
	} else {
		/* new table */
		increase_table ();
		increment_table (k);
		/* send a customer recursively */
		if (parent != NULL)
			return parent->increment_customer (word);
	}

	return true;
}

void
restaurant::increase_table ()
{
	tables.push_back (0);
	context->ntables++;
}

void
restaurant::increment_table (int k)
{
	tables[k]++;
	ncounts++;
	context->ncounts++;
	context->records->incr (this);
}

/* decrement */

bool
restaurant::decrement_customer ()
{
	int k, word_ = word;
	ngram *parent = context->parent;
	ngram *context_ = context;

	if (tables.size() > 1)
	{
		if ((k = draw_discrete_int(&(tables[0]),
					   tables.size())) == -1)
		{
			error("decrement_customer: table invalid.\n");
			return false;
		}
	} else
		k = 0;

	decrement_table(k);
	
	if (tables[k] == 0)	/* table[k] becomes unoccupied */
	{
		decrease_table(k);
		/* remove restaurant if necessary */
		if (tables.size() == 0)
		{
			delete this;
			if (parent != NULL && context_->ntables == 0)
				delete context_;
		}
		if (parent != NULL)
			return parent->decrement_customer (word_);
	}
	
	return true;
}

void
restaurant::decrement_table (int k)
{
	tables[k]--;
	ncounts--;
	context->ncounts--;
	context->records->decr (this);
	if (ncounts == 0)
		context->records->erase (this);
}

void
restaurant::decrease_table (int k)
{
	int i, n = tables.size();

	/* shift tables */
	for (i = k; i < n - 1; i++)
		tables[i] = tables[i + 1];
	/* resize */
	tables.resize(n - 1);
	/* decrement counts */
	context->ntables--;
}

/*
 *  probability
 *
 */

double
ngram::hpy_probability (int w)
{
	int l;
	double c;
	restaurant *rest;

	l = level;

	if ((rest = find_word (words, w)) == NULL)
		/* word not found */
		c = 0;
	else
		/* PY discount */
		c = rest->ncounts - lm->d[l] * rest->tables.size();

	if (parent == NULL)
		return (c + (lm->theta[l] + lm->d[l] * ntables)
			/ lm->lexicon)
			/ (ncounts + lm->theta[l]);
	else
		return (c + (lm->theta[l] + lm->d[l] * ntables)
			* parent->hpy_probability(w))
			/ (ncounts + lm->theta[l]);
}

double
ngram::hpy_residuals ()
{
	int l = level;
	
	return (lm->theta[l] + lm->d[l] * ntables) / (lm->theta[l] + ncounts);
}

/*
 *  misc
 *
 */

static int
print_restaurant (restaurant *rest, int *level)
{
	int i;

	for (i = 0; i < *level; i++) printf("  ");
	printf("restaurant %4d: ncounts = %3d, tables =", rest->word, rest->ncounts);
	for (i = 0; i < rest->tables.size(); i++)
		printf(" %d", rest->tables[i]);
	printf("\n");
	
	return 0;
}

static void
_print_tree (ngram *node, void *arg)
{
	node->print_node ();
	
	if (!node->children->empty())
		hash_foreach (node->children, (nodefunc)_print_tree, NULL);
}

void
ngram::print ()
{
	_print_tree (this, NULL);
}

void
ngram::print_node ()
{
	vector<int>::iterator it;
	
	int i;

	for (i = 0; i < level; i++) printf("  ");
	printf("node: level = %d word = %d, words = %d\n", level, word, words->size());
	for (i = 0; i < level; i++) printf("  ");	
	printf("bounds:");
	for (i = 0; i < records->bounds.size(); i++)
		printf(" %d", records->bounds[i]);
	printf("\n");
	records->foreach ((recfunc)print_restaurant, &level);
	printf("------\n");
}

/*
 *   utility
 *
 */

ngram *
find_node (hashmap *children, int word)
{
	hashmap::iterator it;

	if ((it = children->find (word)) == children->end())
		return NULL;
	else
		return (ngram *) it->second;
}

restaurant *
find_word (hashmap *words, int word)
{
	hashmap::iterator it;

	if ((it = words->find (word)) == words->end())
		return NULL;
	else
		return (restaurant *) it->second;
}

void
insert_node (hashmap *children, ngram *node)
{
	children->insert (cons(node->word, node));
}

void
insert_word (hashmap *words, restaurant *rest)
{
	words->insert (cons(rest->word, rest));
}

void
delete_node (hashmap *children, ngram *node)
{
	hashmap::iterator it;

	if ((it = children->find (node->word)) == children->end())
	{
		error("delete_node: can't delete node %d.\n", node->word);
		return;
	} else {
		children->erase (it);
	}
}

void
delete_word (hashmap *words, restaurant *rest)
{
	hashmap::iterator it;

	if ((it = words->find (rest->word)) == words->end())
	{
		error("delete_word: can't delete word %d.\n", rest->word);
		return;
	} else {
		words->erase (it);
	}
}

