/*
    splay.c
    $Id: splay.c,v 1.1 2010/03/08 08:01:44 daichi Exp $

*/
#include <stdio.h>
#include <stdlib.h>
#include "splay.h"
#define  max(a,b)	  ((a)>(b)?(a):(b))

static splay *
splay_tree (splay *t, void *i, cmpfunc comp)
{
	splay n, *l, *r, *y;
	int z;

	if (t == NULL)  return t;
	n.left = n.right = NULL;
	l = r = &n;

	for (;;)
	{
		z = (*comp)(i, t->item);
		if (z < 0) {
			if (t->left == NULL)  break;
			if ((*comp)(i, t->left->item) < 0) {	/* rotate */
				y = t->left;
				t->left = y->right;
				y->right = t;
				t = y;
				if (t->left == NULL)  break;
			}
			r->left = t;				/* link right */
			r = t;
			t = t->left;
		} else if (z > 0) {
			if (t->right == NULL)  break;
			if ((*comp)(i, t->right->item) > 0) {	/* rotate */
				y = t->right;
				t->right = y->left;
				y->left = t;
				t = y;
				if (t->right == NULL)  break;
			}
			l->right = t;				/* link left */
			l = t;
			t = t->right;
		} else
			break;
	}
	l->right = t->left;
	r->left = t->right;
	t->left = n.right;
	t->right = n.left;
	return t;
}

splay *
splay_insert (splay *t, void *i, cmpfunc comp)
{
	splay *n;
	int z;

	if ((n = (splay *)malloc(sizeof(splay))) == NULL)
	{
		perror("malloc");
		exit(1);
	}
	n->item = i;
	if (t == NULL)
	{
		n->left = n->right = NULL;
		return n;
	}
	t = splay_tree (t, i, comp);
	z = (*comp)(i, t->item);
	if (z < 0) {
		n->left = t->left;
		n->right = t;
		t->left = NULL;
		return n;
	} else if (z > 0) {
		n->right = t->right;
		n->left = t;
		t->right = NULL;
		return n;
	} else {
		/* already in the tree */
		free(n);
		return t;
	}
}

splay *
splay_delete (splay *t, void *i, cmpfunc comp)
{
	splay *x;

	if (t == NULL)  return NULL;
	t = splay_tree (t, i, comp);
	if ((*comp)(i, t->item) == 0)
	{
		/* found */
		if (t->left == NULL) {
			x = t->right;
		} else {
			x = splay_tree (t->left, i, comp);
			x->right = t->right;
		}
		free(t);
		return x;
	}
	/* not found */
	return t;
}

void *
splay_find (splay **tp, void *i, cmpfunc comp)	/* NULL -> fail, else success */
{
	if (*tp == NULL) return NULL;
	*tp = splay_tree (*tp, i, comp);
	if ((*comp)(i, (*tp)->item) == 0)	/* found */
		return (*tp)->item;
	else
		return NULL;
}

void *
splay_lookup (splay *t, void *i, cmpfunc comp)	/* without tree update */
{
	int z;
	
	if (t == NULL)
		return NULL;
	z = (*comp)(i, t->item);
	if (z < 0)
		return splay_lookup (t->left, i, comp);
	else if (z > 0)
		return splay_lookup (t->right, i, comp);
	else
		return t->item;
}

splay *
splay_free (splay *t)
{
	if (t == NULL)
		return NULL;
	else {
		splay_free (t->left);
		splay_free (t->right);
		free(t);
		return NULL;
	}
}

void
splay_foreach (splay *t, nodefunc func, void *arg)
{
	if (t == NULL)
		return;
	else {
		splay_foreach (t->left, func, arg);
		(*func)(t->item, arg);
		splay_foreach (t->right, func, arg);
	}
}

double
splay_fold (splay *t, foldfunc func, void *arg)
{
	if (t == NULL)
		return 0;
	else {
		return (*func)(t->item, arg)
			+ splay_fold (t->left, func, arg)
			+ splay_fold (t->right, func, arg);
	}
}

int
splay_max (splay *t, maxfunc func, void *arg)
{
	int x, y, z;
	
	if (t == NULL)
		return 0;
	else {
		x = (*func)(t->item, arg);
		y = splay_max (t->left, func, arg);
		z = splay_max (t->right, func, arg);
		return max (x, max (y, z));
	}
}

int
splay_nodes (splay *t)
{
	if (t == NULL)
		return 0;
	else
		return 1 + splay_nodes(t->left)
			 + splay_nodes(t->right);
}

static void
_splay_print (splay *t, prfunc print, int n)
{
	int i;
	
	if (t == NULL)
		return;
	else {
		_splay_print (t->left, print, n + 1);
		for (i = 0; i < n; i++) printf("  ");
		(*print)(t->item);
		_splay_print (t->right, print, n + 1);
	}
}

void
splay_print (splay *t, prfunc print)
{
	_splay_print(t, print, 0);
}

