/*
    table.cpp
    Lexical translation sparse table.
    $Id: table.cpp,v 1.9 2010/03/17 16:41:57 daichi Exp $

*/
#include "lwlm.h"
#include "util.h"
#include <ext/hash_map>

using namespace std;
using namespace __gnu_cxx;

TransTable::TransTable (double alpha_)
{
	alpha = alpha_;
}

bool
TransTable::init (int lexicon)
{
	L = lexicon;
	total.resize (L);
	fill (total.begin(), total.end(), 0);
}

int
TransTable::count (int hidden, int observed)
{
	val_hash::iterator it;
	long idx = hidden * (long) L + observed;

	if ((it = counts.find(idx)) == counts.end())
	{
		return -1;
	} else {
		return it->second;
	}
}

void
TransTable::add_count (int hidden, int observed)
{
	val_hash::iterator it;
	long idx = hidden * (long) L + observed;

	if ((it = counts.find(idx)) == counts.end())
	{
		counts.insert (cons(idx, 1));
	} else {
		it->second++;
	}
	total[hidden]++;
}

void
TransTable::sub_count (int hidden, int observed)
{
	val_hash::iterator it;
	long idx = hidden * (long) L + observed;

	if ((it = counts.find(idx)) == counts.end())
	{
		error("sub_count: can't find a value for %d -> %d.\n",
		      hidden, observed);
		return;
	} else {
		it->second--;
		if (it->second == 0)
		{
			counts.erase (it);
		}
		total[hidden]--;
	}
}

double
TransTable::probability (int hidden, int observed)
{
	long idx;
	double c;
	val_hash::iterator it;

	if (hidden > L || observed > L)
		return alpha / (alpha * L);	/* for decoding */

	idx = hidden * (long) L + observed;
	
	if ((it = counts.find(idx)) == counts.end())
	{
		c = alpha;
	} else {
		c = it->second + alpha;
	}

	return c / (total[hidden] + alpha * L);
}

bool
TransTable::save (char *file, LWLM *lwlm)
{
	val_hash::iterator it;
	int hidden, observed;
	FILE *fp;

	if ((fp = fopen(strconcat(file, "/table"), "w")) == NULL)
	{
		perror("fopen");
		return false;
	}

	for (it = counts.begin(); it != counts.end(); it++)
	{
		hidden = it->first / L;
		observed = it->first % L;

		fprintf(fp, "%s\t%s\t%d\t%.7f\n", lwlm->ids(hidden), lwlm->ids(observed),
			it->second, probability(hidden, observed));
	}

	fclose(fp);
}

bool
TransTable::load (char *file, LWLM *lwlm)
{
	FILE *fp;
	double p;
	int  i, h, o, count;
	char buf [BUFSIZ];
	char hidden [BUFSIZ], observed [BUFSIZ];

	if ((fp = fopen(strconcat(file, "/table"), "r")) == NULL)
	{
		perror("fopen");
		return false;
	}

	while (fgets(buf, sizeof(buf), fp))
	{
		if (sscanf(buf, "%s\t%s\t%d\t%lg",
			   hidden, observed, &count, &p) != 4)
		{
			error("TransTable::load: invalid line.\n");
			return false;
		}
		h = lwlm->sid (hidden);
		o = lwlm->sid (observed);

		for (i = 0; i < count; i++)
			add_count (h, o);
	}

	fclose(fp);
}
	

#if 0
#include <cstdio>

int
main (int argc, char *argv[])
{
	TransTable *table = new TransTable();
	int i, j, L = 10;
	double p, z;

	table->init(L, 0.2);

	table->add_count(3, 4);
	table->add_count(2, 9);
	table->add_count(3, 4);	
	table->sub_count(2, 9);

	for (i = 0; i < L; i++)
	{
		z = 0;
		for (j = 0; j < L; j++)
		{
			p = table->probability(i, j);
			printf("p(%d|%d) = %g\n", j, i, p);
			z += p;
		}
		printf("total[%d] = %d\n", i, table->total[i]);
		printf("p(*|%d) = %g\n", i, z);
	}

}

#endif

		
